function res = class_CVdemo2(opts,pars)
% class_CVdemo2: demonstrates cross-validated classification with SFA
%
% see [Berkes05] Pietro Berkes: Pattern recognition with Slow Feature
% Analysis. Cognitive Sciences EPrint Archive (CogPrint) 4104, http://cogprints.org/4104/ (2005)
%
% Instead of handwritten digits we use: 
% - with pars.GENERATE_P==2 or ==3 the Vowel benchmark dataset from UCI repository
% http://archive.ics.uci.edu/ml/datasets/Connectionist+Bench+%28Vowel+Recognition+-+Deterding+Data%29 
% - or with pars.GENERATE_P==4 the Breast Cancer dataset from UCI repository
% http://archive.ics.uci.edu/ml/datasets/Breast+Cancer+Wisconsin+%28Diagnostic%29
% - or with pars.GENERATE_P==9 the gesture data from Polak/Hein (FH Cologne).
% - or with pars.GENERATE_P==10 the new gesture data from Kristine Hein (FH Cologne).
%
% We do pars.nfold cross-validation.
%
% INPUT
%   opts        struct with several parameter settings for CLASSIFY. If omitted on
%               input, meaningful defaults will be set (see source code)
%   pars        struct with several parameter settings for training and test data generation. 
%               If omitted on input, meaningful defaults will be set (see source code)
% OUTPUT
%   res         struct with the results from the *last* classify-run (fields
%               y, ytst, predT, predX, see classify.m) plus the fields
%       errtrn  1 x 2 matrix: error rate with / w/o SFA on training set (mean over all CV-runs)
%       errtst  1 x 2 matrix: error rate with / w/o SFA on test set (mean over all CV-runs)
%       allerrtrn   nfold x 2 matrix (results from each CV run)
%       allerrtst   nfold x 2 matrix (results from each CV run)
%       realclass   nrecord x 1: the true class for each record
%       cvpredict   nrecord x 1: the CV-predicted class for each record
%       confmat     the confusion matrix of CV-predictions
%
% Example settings for struct pars:
%       {GENERATE_P=10; nfold=10; tstvarseed=0; ncopies=0;}
%
% Author: Wolfgang Konen, Sep 2009
%
% See also: CLASSIFY, CLASS_DEMO2

    if nargin<2
        pars.GENERATE_P=10;  % =2: load 'vowel-context.data' (UCI repository)  
                             % =3: load 'vowel-train.csv', 'vowel-test.csv' (Renal assignment)
                             % =4: load 'wdbc.data (UCI repository, malignant & benign breast cancer)
                             % =9: load 'class_pat_gest.mat' (gesture data Polak/Hein)
                             % =10: load 'hk_gesture_all-norm*.mat' (new gesture data K.Hein) 
        pars.tstvarseed=0;   % =1: vary seed, =0: always seed 0, for random number generator
                             % in case of cross validation index 
        pars.nfold=10;       % number of folds in cross validation index
        pars.ncopies=0;      % if >0, add pars.ncopies noisy copies of the training data to
                             % the training data (this is for the case of few patterns per
                             % class, as compared to idim, to avoid overfitting)
        pars.PERSON_ID= -1;  % -1: all persons, 121...130: only specific person
                             % (only for pars.GENERATE_P==10)
    end
    if nargin<1
        opts.ALIGNED=1;      % =1: align the Gaussian classifiers with coordinate axes 
                             % (recommended, this is more robust if the data distribution deviates
                             % largely from multivariate Gaussian). 
                             % =0: don't align, can be more precise if data are Gaussian. 
        opts.dographics=2;   % =0: no plots, =1: some plots, =2: more plots
        opts.dbg=0;          % =1 for activating sfa_dbg in sfa2_step, else =0  
        opts.epsC = 0;       % cutoff eigenvalues of C smaller than epsC*lammax (def. 1e-7)   
        %
        % some more fields of struct opts will be set after loading the data
        %   opts.idim       % number of input dimensions
        %   opts.pp_range   % dims after preproc
        %   opts.classes    % labels of the classes (we assume that each class appears
                            % at least once in the training data)
        %   opts.nclass     % number of classes
        %   opts.gaussdim   % number of SFA-outputs used in Gauss classifier 
    end

    if pars.GENERATE_P==2
        % http://archive.ics.uci.edu/ml/datasets/Connectionist+Bench+%28Vow
        % el+Recognition+-+Deterding+Data%29  (UCI repository, vowel)
        M0=dlmread('data/vowel-context.data');   
        % IMPORTANT: let DLMREAD infer the delimeter (one or more
        % whitespaces) from file, because only in this way, consecutive
        % whitespaces are treated as ONE delimiter
        %
        M=M0(:,4:14);
        c_col = 11;         % class column
        i_col = 1:10;       % input columns range
    elseif pars.GENERATE_P==3
        % read Vowel data from
        % http://www.inf.ed.ac.uk/teaching/courses/inf2b/coursework/cwk3.html 
        % (Steve Renals Homework Assignment) 
        % Same IO-records as in GENERATE_P==2, but different divide in train & test
        %
        Mtrn=csvread('data/vowel-train.csv');
        Mtst=csvread('data/vowel-test.csv');
        M = [Mtrn; Mtst];
        c_col = 11;         % class column
        i_col = 1:10;       % input columns range
    elseif pars.GENERATE_P==4
        % http://archive.ics.uci.edu/ml/datasets/Breast+Cancer+Wisconsin+(Diagnostic)
        % UCI repository, breast cancer, with replacement in column 2:
        %   M >> -1 (malignant)   and     B >> 2 (benign)
        M=csvread('data/wdbc.data');   
        c_col = 2;          % class column
        i_col = 3:32;       % input columns range
        %M(:,i_col)=M(:,i_col)./repmat(mean(M(:,i_col)),size(M,1),1);
        parallel_plot(M(:,i_col),M(:,c_col));
    elseif pars.GENERATE_P==9 
        % read gesture data 
        % as they were generated by beginner.m (Marco Polak, Kristine Hein)
        load('data/class_pat_gest.mat');
        i_col = 2:31;       % input columns range        
        M=[realclass' x];
        c_col = 1;              % class column
        opts.pp_range=11;
        parallel_plot(x,realclass);              
    elseif pars.GENERATE_P==10
        % read gesture data 
        % as they were generated load_gesturedata2.m (Kristine Hein, WK)
        load('data/hk_gesture_all-norm1.mat');     % all 716 patterns
        %load('data/hk_gesture_p130.mat');          % only person 130
        if pars.PERSON_ID~=-1
            idx = find(info_preproc(:,1)==pars.PERSON_ID);
            realclass = realclass(idx);
            x = x(idx,:);
            info_preproc = info_preproc(idx,:);
            disp(['Selected subset of data for PERSON_ID=' num2str(pars.PERSON_ID);]);
        end
        imax = size(x,2);       % 91 for -norm0, 92 for -norm1
        i_col = 2:imax;         % input columns range        
        c_col = 1;              % class column
        M=[realclass' x];
        opts.pp_range=11;
        parallel_plot(x,realclass);              
        cvi = info_preproc(:,1)'-120;   % optional: cross validation over different persons 
    end % pars.GENERATE_P
    
    opts.classes = unique(M(:,c_col));  % labels of the classes 
    opts.nclass=numel(opts.classes);    % number of classes
    opts.gaussdim=opts.nclass-1;        % number of SFA-outputs used in Gauss classifier
    opts.idim=numel(i_col);             % number of input dimensions
    if ~isfield(opts,'pp_range')
        opts.pp_range=opts.idim;        % dims after preproc = number of input dims
    end
    %if pars.GENERATE_P==4
    %    opts.gaussdim=3;                % do not use only the slowest SFA output (because nclass=2)
    %end                                 % but use the slowest 3 SFA outputs        
    
    % create crossvalidation index (if not done before, see pars.GENERATE_P==10)
    if ~exist('cvi','var')
        cvi = cvi_create(pars,size(M,1));   % cross validation over pars.nfold random subsets 
    end

    % only needed for comparision with R and RF:
    % (see <fiwa_soma>\src\trunk\R-DM-Template\ClassifyTemplate\main_CV_gesture.r)
    write_gestureCV_csv(cvi,M);
    
    %
    % run the pars.nfold=10 folds through SFA
    % 
    allerrtrn = [];
    allerrtst = [];
    cases_trn = [];
    cases_tst = [];
    allrealclass = M(:,c_col);
    allcvpredict =0*M(:,c_col);
    pclasserr = [];             % class errors for each fold
    for i=1:pars.nfold
        idxtst = find(cvi==i);
        idxtrn = find(cvi~=i);
        x    = M(idxtrn,i_col);
        xtst = M(idxtst,i_col);
        realclass = M(idxtrn,c_col)';    
        realc_tst = M(idxtst,c_col)';   

        % add pars.ncopies noisy copies, but only to the training data
        [realclass, x] = add_noisy_copies(realclass,x,pars.ncopies);

        %
        % the main call to SFA classification
        %
        res = classify(x,realclass,xtst,realc_tst,opts);
        
        allerrtrn = [allerrtrn; res.errtrn];
        allerrtst = [allerrtst; res.errtst];
        cases_trn = [cases_trn; size(x,1)];
        cases_tst = [cases_tst; size(xtst,1)];
        allcvpredict(idxtst) = res.predT;
        [confmat,classerr] = mk_confmat(allrealclass(idxtst),res.predT,opts.nclass)
        pclasserr = [pclasserr; classerr];
        i
    end
    
    [confmat,classerr] = mk_confmat(allrealclass,allcvpredict,opts.nclass);
    fprintf('Confusion matrix & class errors in %%\n    --- true class ---\n')
    disp(confmat);
    disp(['  ' num2str(classerr(1:opts.nclass)*100,'%6.1f') ...
          '%   Total: ' num2str(classerr(opts.nclass+1)*100,'%6.2f%%')]);
    
    res.allerrtrn = allerrtrn;
    res.allerrtst = allerrtst;
    w_trn = cases_trn/sum(cases_trn);
    w_tst = cases_tst/sum(cases_tst);
    if size(allerrtrn,1)>1
        res.errtrn = sum(repmat(w_trn,1,2) .* allerrtrn);
        res.errtst = sum(repmat(w_tst,1,2) .* allerrtst);
    end
    res.realclass = allrealclass;
    res.cvpredict = allcvpredict;
    res.confmat = confmat;
    res.pclasserr = pclasserr;
    
  
    fprintf('                                with SFA   w/o SFA\n');    
    fprintf('CV error rate on training set = %g%%       %g%%\n', res.errtrn*100);    
    fprintf('CV error rate on     test set = %g%%       %g%%\n', res.errtst*100); 
    %numel(find(allrealclass~=allcvpredict))/numel(allrealclass)

end 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% some private helper functions
%
function cvi = cvi_create(pars,sz)
    %
    % create the crossvalidation index CVI with pars.nfold folds
    %
    pars.nfold = 10.;
    cvi = zeros(1,sz);
    ilow=1;
    for i=1:pars.nfold
        ihigh = fix(sz/pars.nfold*i);
        cvi(ilow:ihigh) = i;        % i.e. [111..2222...333...]
        ilow = ihigh+1;
    end
    if pars.tstvarseed==0
        rand('state',4);               % each time the same random numbers
    else
        rand('state',sum(100*clock));   % each time different random numbers
    end    
    cvi = cvi(randperm(sz));        % permute CVI randomly 
end

function write_gestureCV_csv(cvi,M)
    % write the matrix of exactly this CV-division of
    % all data on file for comparision with R and RF (random
    % forest). First column: CV index.
    path='C:\WUTemp\FH-MassenDaten\fiwa_soma\src\trunk\R-DM-Template\ClassifyTemplate\data\';
    %path='';
    dlmwrite([path 'gestureCV_pat.csv'],[cvi' M],';');
    fprintf('Gesture data written onto gestureCV_pat.csv\n');
end 

function [realclass, x] = add_noisy_copies(realclass,x,ncopies)
    if ncopies>0
        % add noisy copies, but only to the training data
        x0 = x;
        for k=1:ncopies
            x2 = x0 + 0.1*repmat(std(x0),size(x0,1),1).*randn(size(x0,1),size(x0,2));
            x = [x; x2];
        end
        realclass = repmat(realclass,1,ncopies+1);
        % sort the training data class-by-class
        [tmp,idx] = sort(realclass);
        realclass = realclass(idx);
        x = x(idx,:);
    end
end
