function res = class_demo1(opts,pars)
% class_demo1: demonstrate how classification can be done with SFA
%
% see [Berkes05] Pietro Berkes: Pattern recognition with Slow Feature
% Analysis. Cognitive Sciences EPrint Archive (CogPrint) 4104, http://cogprints.org/4104/ (2005)
%
% Instead of handwritten digits we generate artificial (synthetic) classification data 
%
% INPUT
%   opts        struct with several parameter settings for CLASSIFY. If omitted on
%               input, meaningful defaults will be set (see source code)
%   pars        struct with several parameter settings for training and test data generation. 
%               If omitted on input, meaningful defaults will be set (see source code)
% OUTPUT
%   res         struct with fields
%       errtrn  1 x 2 matrix: error rate with / w/o SFA on training set 
%       errtst  1 x 2 matrix: error rate with / w/o SFA on test set 
%
% Author: Wolfgang Konen, Sep 2009
%
% See also: CLASS_DEMO2, CLASSIFY, CLASS_SWEEP

    if nargin<2
        pars.npatt=100;      % number of patterns per class
        pars.sigma=0.08;     % sigma of Gaussian noise added to prototypes
        pars.GENERATE_P=5;   % =1: generate synthetic class patterns x (& save on data/class_pat.mat)
                             % =0: load synthetic class patterns x from data/class_pat.mat  
                             % =5: generate totally random synthetic data
        pars.MIXING=1;       % =1: nonlinear mixing of input variables, same for all classes
                             % =2: nonlinear mixing in x1 and x2, different for all classes
                             % other: no mixing
        pars.tstfrac=0.40;   % which fraction of the patterns becomes test patterns
        pars.tstvarseed=1;   % =1: vary seed, =0: always seed 0, for random number generator
                             % in case of random divide
    end
    if nargin<1
        opts.idim=5;         % number of input dimensions
        opts.nclass=4;       % number of classes
        opts.ALIGNED=1;      % =1: align the Gaussian classifiers with coordinate axes 
                             % (recommended, this is more robust if the data distribution deviates
                             % largely from multivariate Gaussian). 
                             % =0: don't align, can be more precise if data are Gaussian. 
        opts.dographics=2;   % =0: no plots, =1: some plots, =2: more plots
        opts.dbg=0;          % =1 for activating sfa_dbg in sfa2_step, else =0  
        opts.epsC=0;
    end
    
    if pars.GENERATE_P==5
        % totaly random data (points taken from the opts.idim-dimensional
        % [0,1]-cube) --> nothing can be learned
        pars.npatt=30;
        opts.idim=20;
        opts.nclass=5;
        opts.pp_range=opts.idim;            % number of preprocessing dimensions     
        %opts.pp_range=5;
        realclass = reshape(repmat([1:opts.nclass], pars.npatt,1),1,opts.nclass*pars.npatt);   % [111..222..333....]
        x = rand(pars.npatt*opts.nclass,opts.idim);
    end

    idim    =opts.idim;
    nclass  =opts.nclass;
    npatt   =pars.npatt;

    if pars.GENERATE_P<=1
        if pars.GENERATE_P==1
            % for the very first time one pass with GENERATE_P==1 is necessary
            %
            prototype = 0.2+rand(nclass,idim);  % prototype(n,:) is prototype vector for class n
            x=[];
            for n=1:nclass
                x = [x; repmat(prototype(n,:),npatt,1)+randn(npatt,idim)*pars.sigma ];
            end
            x(:,5:idim) = rand(size(x(:,5:idim)));    %distractive dimension(s)
            realclass = reshape(repmat([1:nclass], npatt,1),1,nclass*npatt);   % [111..222..333....]
            save('data/class_pat.mat','x','realclass');
        elseif pars.GENERATE_P==0    
            load('dtaa/class_pat.mat');
        end 

        if pars.MIXING==1
            % make the class separation a little harder in input space:
            if idim>2
                x(:,1)=x(:,1)./x(:,3);
                x(:,2)=x(:,2)./x(:,3);
            end
            if idim>3
                x(:,2)=x(:,2)./x(:,4);
                x(:,3)=x(:,3)./x(:,4);
            end
        elseif pars.MIXING==2
            % make the class separation considerably harder in input space:
            for n=1:nclass
                m=mod(n,idim)+1;
                n2=mod(n+2,idim)+1;
                m2=mod(n2,idim)+1;
                ind=find(realclass==n);
                x(ind,n)=rand(numel(ind),1);
                x(ind,m)=rand()+x(ind,n).^2*rand()+rand(numel(ind),1)*pars.sigma;
                x(ind,n2)=rand(numel(ind),1);
                x(ind,m2)=rand()+x(ind,n2).^2*rand()+rand(numel(ind),1)*pars.sigma;
            end
        end % pars.MIXING
    end

    [x,xtst,realclass,realc_tst] = divide_rand(x,realclass,pars);

    opts.gaussdim=opts.nclass-1;        % number of SFA-outputs used in Gauss classifier    
    opts.classes = unique(realclass);   % labels of the classes (we assume that each class appears
                                        % at least once in the training data)
    if ~isfield(opts,'pp_range')
        opts.pp_range=opts.idim;        % dims after preproc = number of input dims
    end

    res = classify(x,realclass,xtst,realc_tst,opts);
end 

function [x,xtst,realclass,realc_tst] = divide_rand(x,realclass,pars)
    % divide x and realclass in test and training patterns by random split
    if pars.tstvarseed==0
        rand('state',0);               % each time the same random numbers
    else
        rand('state',sum(100*clock));   % each time different random numbers
    end
    sz=size(x,1);
    tstind=zeros(1,sz);
    tstind(1,rand(1,sz)<pars.tstfrac) = 1;
    xtst = x(tstind==1,:);
    x    = x(tstind==0,:);
    realc_tst = realclass(tstind==1);
    realclass = realclass(tstind==0);
end    
    
