function res = class_demo2(opts,pars)
% class_demo2: demonstrate how classification can be done with SFA
%
% see [Berkes05] Pietro Berkes: Pattern recognition with Slow Feature
% Analysis. Cognitive Sciences EPrint Archive (CogPrint) 4104, http://cogprints.org/4104/ (2005)
%
% Instead of handwritten digits we use the Vowel or WDBC benchmark dataset 
% from UCI machine learning repository (http://archive.ics.uci.edu/ml) 
% or gesture data (Polak/Hein, FH Kln) from Cologne University
%
% INPUT
%   opts        struct with several parameter settings for CLASSIFY. If omitted on
%               input, meaningful defaults will be set (see source code)
%   pars        struct with several parameter settings for training and test data generation. 
%               If omitted on input, meaningful defaults will be set (see source code)
% OUTPUT
%   res         struct with the results from classify-run, e.g. 
%       errtrn  1 x 2 matrix: error rate with / w/o SFA on training set 
%       errtst  1 x 2 matrix: error rate with / w/o SFA on test set 
%       y       output from SFA when applied to training data
%       ytst    output from SFA when applied to test data
%       predT   predictions with SFA + GaussClassifier on test set
%       predX   predictions w/o SFA (only GaussClassifier) on test set 
%
% Example settings for struct pars:
%       {GENERATE_P=2; tstdiv=0; tstfrac=0.1; tstvarseed=0; ncopies=0;}
%       {GENERATE_P=9; tstdiv=1; tstfrac=0.1; tstvarseed=0; ncopies=1;}
%       {GENERATE_P=10; tstdiv=1; tstfrac=0.1; tstvarseed=0; ncopies=1;}
%
% Author: Wolfgang Konen, May 2009 - Nov 2009
%
% See also: CLASSIFY, CLASS_SWEEP, CLASS_CVDEMO2

    if nargin<2
        pars.GENERATE_P=10;  % =1: load 'Vowel.csv' (R mlbench), divide randomly in tst/trn
                             % =2: load 'vowel-context.data' (UCI repository, recommended)  
                             % =3: load 'vowel-train.csv', 'vowel-test.csv' (Renal assignment)
                             % =4: load 'wdbc.data (UCI repository, malignant & benign breast cancer)
                             % =9: load 'class_pat_gest.mat' (gesture data Polak/Hein)
                             % =10: load 'hk_gesture_all-norm*.mat' (new gesture data K.Hein) 
        pars.tstdiv=1;       % =1: divide randomly in tst/trn even in case GENERATE_P=2 or 3
                             % =0: keep tst/trn from file in case GENERATE_P=2 or =3
        pars.tstfrac=0.10;   % which fraction of the patterns become test 
                             % patterns in case of random divide
        pars.tstvarseed=1;   % =1: vary seed, =0: always seed 0, for random number generator
                             % in case of random divide
        pars.ncopies=0;      % if >0, add pars.ncopies noisy copies of the training data to
                             % the training data (this is for the case of few patterns per
                             % class, as compared to idim, to avoid overfitting)
        pars.PERSON_ID= -1;  % -1: all persons, 121...130: only specific person
                             % (only for pars.GENERATE_P==10)
    end
    if nargin<1
        opts.ALIGNED=1;      % =1: align the Gaussian classifiers with coordinate axes 
                             % (recommended, this is more robust if the data distribution deviates
                             % largely from multivariate Gaussian). 
                             % =0: don't align, can be more precise if data are Gaussian. 
        opts.dographics=3;   % =0: no plots, =1: some plots, =2: more plots
        opts.dbg = 1;        % =1 for activating sfa_dbg in sfa2_step, else =0   
        opts.epsC = 0;       % cutoff eigenvalues of C smaller than epsC*lammax (def. 1e-7)   

        % some more fields of struct opts will be set after loading the data:
        %   opts.idim       % number of input dimensions
        %   opts.pp_range   % dims after preproc
        %   opts.classes    % labels of the classes (we assume that each class appears
                            % at least once in the training data)
        %   opts.nclass     % number of classes
        %   opts.gaussdim   % number of SFA-outputs used in Gauss classifier 
                                        
    end
    
    DBG=0;
    if pars.GENERATE_P==1 
        % read Vowel data (see R-Project, help on Vowel(mlbench), the CSV was written from R with
        %   library(mlbench)
        %   data(Vowel)
        %	V2 = cbind(Vowel, as.numeric(Vowel$Class))
        %   write.table(V2[,setdiff(names(V2,Class], file=Vowel.csv, quote=F, sep=";", dec=".", row.names=F, col.names=T)
        % ) 
        % *** CAUTION, NOT RECOMMENDED ***, since the R-dataset seems incomplete
        % (one input dim missing and not the right train/test division)
        % This branch resides only in here to reproduce the data error
        M=dlmread('data/Vowel.csv',';',1,0);
        x=M(:,2:10);
        realclass = M(:,11)';

        [x,xtst,realclass,realc_tst] = divide_rand(x,realclass,pars);
    end % pars.GENERATE_P==2

    if pars.GENERATE_P==2  || DBG==1
        % read Vowel data from
        % http://archive.ics.uci.edu/ml/datasets/Connectionist+Bench+%28Vow
        % el+Recognition+-+Deterding+Data%29  (UCI repository)
        M=dlmread('data/vowel-context.data');   
        % IMPORTANT: let DLMREAD infer the delimeter (one or more
        % whitespaces) from file, because only in this way, consecutive
        % whitespaces are treated as ONE delimiter
        %
        idx=find(M(:,1)==0);            % training index
        x=M(idx,4:13);
        realclass = M(idx,14)';    
        parallel_plot(x,realclass);
        idx=find(M(:,1)==1);            % test index
        xtst=M(idx,4:13);
        realc_tst = M(idx,14)';   
        
        if pars.tstdiv==1
            x = M(:,4:13);
            realclass = M(:,14)';
            [x,xtst,realclass,realc_tst] = divide_rand(x,realclass,pars);
        end

    end 
    
    if pars.GENERATE_P==3
        % read Vowel data from
        % http://www.inf.ed.ac.uk/teaching/courses/inf2b/coursework/cwk3.html 
        % (Steve Renals Homework Assignment) 
        % Same IO-records as in GENERATE_P==2, but different divide in train & test
        %
        Mtrn=csvread('data/vowel-train.csv');
        x=Mtrn(:,1:10);
        realclass = Mtrn(:,11)';    

        Mtst=csvread('data/vowel-test.csv');
        xtst=Mtst(:,1:10);
        realc_tst = Mtst(:,11)';
    
        if pars.tstdiv==1
            x = [Mtrn(:,1:10); Mtst(:,1:10)];
            realclass = [Mtrn(:,11); Mtst(:,11)]';
            [x,xtst,realclass,realc_tst] = divide_rand(x,realclass,pars);
        end
        %
    elseif pars.GENERATE_P==4
        % http://archive.ics.uci.edu/ml/datasets/Breast+Cancer+Wisconsin+(Diagnostic)
        % UCI repository, breast cancer, with replacement in column 2:
        %   M >> -1 (malignant)   and     B >> 2 (benign)
        M=csvread('data/wdbc.data');   
        c_col = 2;          % class column
        i_col = 3:32;       % input columns range
        x = M(:,i_col);
        [x,xtst,realclass,realc_tst] = divide_rand(x,M(:,c_col)',pars);
        parallel_plot(M(:,i_col),M(:,c_col));
        %
    elseif pars.GENERATE_P==9 | pars.GENERATE_P==10
        % read gesture data 
        % as they were generated by beginner.m (Marco Polak, Kristine Hein)
        % or by load_gesturedata2.m (Kristine Hein, WK)
        %
        if (pars.GENERATE_P==9) 
            load('data/class_pat_gest.mat');
        else % i.e. if pars.GENERATE_P==10
            load('data/hk_gesture_all-norm1.mat');
        end    
        if pars.PERSON_ID~=-1
            %idx = find(info_preproc(:,1)==pars.PERSON_ID);
            idx = find(info_preproc(:,1)==pars.PERSON_ID & realclass'~=1);
            realclass = realclass(idx);
            x = x(idx,:);
            info_preproc = info_preproc(idx,:);
            disp(['Selected subset of data for PERSON_ID=' num2str(pars.PERSON_ID);]);
        end
        opts.pp_range=11;
        parallel_plot(x,realclass);
        
        [x,xtst,realclass,realc_tst] = divide_rand(x,realclass,pars);
       
        %x0 = x;
        %x = x0 + 0.1*repmat(std(x0),size(x0,1),1).*randn(size(x0,1),size(x0,2));
        if pars.ncopies>0
            x0 = x;
            for i=1:pars.ncopies
                x2 = x0 + 0.1*repmat(std(x0),size(x0,1),1).*randn(size(x0,1),size(x0,2));
                x = [x; x2];
            end
            realclass = repmat(realclass,1,pars.ncopies+1);
            [tmp,idx] = sort(realclass);
            realclass = realclass(idx);
            x = x(idx,:);
        end

        % only needed for comparision with R and RF:
        % (see <fiwa_soma>\src\trunk\R-DM-Template\ClassifyTemplate\main_gesture.r)
        write_gesture_csv(x,realclass,xtst,realc_tst);
    end % pars.GENERATE_
    
    if DBG==1 && pars.GENERATE_P==3, dbg_P2_P3(M,Mtrn,Mtst); end
    
    opts.classes = unique(realclass);   % labels of the classes (we assume that each class appears
                                        % at least once in the training data)
    opts.nclass=numel(opts.classes);    % number of classes
    opts.gaussdim=opts.nclass-1;        % number of SFA-outputs used in Gauss classifier    
    opts.idim=size(x,2);                % number of input dimensions
    if ~isfield(opts,'pp_range')
        opts.pp_range=opts.idim;        % dims after preproc = number of input dims
    end
                                        
    res = classify(x,realclass,xtst,realc_tst,opts);

    % confusion matrix & class error on test set
    [confmat,classerr] = mk_confmat(realc_tst,res.predT,opts.nclass);
    fprintf('Confusion matrix & class errors on test set\n    --- true class ---\n')
    disp(confmat);
    disp(num2str(classerr,'%7.3f'));
    
end 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% some private helper functions
%
function [x,xtst,realclass,realc_tst] = divide_rand(x,realclass,pars)
    % divide x and realclass in test and training patterns by random split
    if pars.tstvarseed==0
        rand('state',4);               % each time the same random numbers
    else
        rand('state',sum(100*clock));   % each time different random numbers
    end
    sz=size(x,1);
    tstind=zeros(1,sz);
    tstind(1,rand(1,sz)<pars.tstfrac) = 1;
    xtst = x(tstind==1,:);
    x    = x(tstind==0,:);
    realc_tst = realclass(tstind==1);
    realclass = realclass(tstind==0);
end    

function dbg_P2_P3(M,Mtrn,Mtst)
    % check whether the records are the same with GENERATE_P==2 or ==3 
    % Answer: they are, but they have different ordering. If we order
    % each input column ascendingly, then each ordered column has
    % exactly the same contents.
    for k=0:9
        [tmp,idx1]=sort(M(:,4+k));
        M1 = M(idx1,:);
        M3 = [Mtrn(:,1:10); Mtst(:,1:10)];
        [tmp,idx3]=sort(M3(:,1+k));
        M3 = M3(idx3,:);
        Mdiff = abs(M1(:,4+k)-M3(:,1+k));
        fprintf('max(abs(M2-Mtrn)) = %g (col %d)\n',max(max(Mdiff)),k);
    end
end

function write_gesture_csv(x,realclass,xtst,realc_tst) 
    % write the matrix of exactly this test-training-division of
    % all data on file for comparision with R and RF (random
    % forest). First column: 1/0 for test/train record.
    tstidx = repmat(1,numel(realc_tst),1); 
    M = [tstidx realc_tst' xtst];  
    tstidx = repmat(0,numel(realclass),1);
    M = [M; tstidx realclass' x];
    path='data/';
    %path='C:\WUTemp\FH-MassenDaten\fiwa_soma\src\trunk\R-DM-Template\ClassifyTemplate\data\';
    dlmwrite([path 'gesture_pat.csv'],M,';');
    fprintf('Gesture data written onto gesture_pat.csv\n');
end 
