function res = classify(x,realclass,xtst,realc_tst,opts)
% classify: demonstrate how classification can be done with SFA
%
% see [Berkes05] Pietro Berkes: Pattern recognition with Slow Feature
% Analysis. Cognitive Sciences EPrint Archive (CogPrint) 4104, http://cogprints.org/4104/ (2005)
%
% INPUT
%   x           NREC x IDIM, training input data
%   xtst        NREC x IDIM, test input data
%   realclass   1 x NREC, training class labels
%   realc_tst   1 x NREC, test class labels
%   opts        struct with several parameter settings. 
%
% OUTPUT
%   res         struct with fields
%       errtrn  1 x 2 matrix: error rate with / w/o SFA on training set 
%       errtst  1 x 2 matrix: error rate with / w/o SFA on test set 
%       y       output from SFA when applied to training data
%       ytst    output from SFA when applied to test data
%       predT   predictions with SFA + GaussClassifier on test set
%       predX   predictions w/o SFA (only GaussClassifier) on test set 
%
% Author: Wolfgang Konen, May 2009 - Aug 2009
%
% See also: CLASS_DEMO1, CLASS_DEMO2, CLASS_CVDEMO2, CLASS_SWEEP

idim    =opts.idim;
nclass  =opts.nclass;

%sfa_clearall;

trnnum=numel(realclass);        % number of training cases

if opts.dographics>0
    cmap = colormap(jet);
    colors=cmap(mod(16*(1:nclass),size(cmap,1))+1,:); 
    figure(1); clf; set(gcf, 'Position', [89 77 866 638]);
    spl = 0;    % subplot counter
end
if opts.dographics>=2
    % scatter plot input signal (first 2 dims, all classes in different
    % colors)    
    spl=spl+1; subplot(2,3,spl);   
    hold on; 
    for i=1:nclass
        ind = find(realclass==opts.classes(i));
        plot(x(ind,1),x(ind,2),'Marker','.','MarkerSize',5,'Color',colors(i,:),'LineStyle','none');
        xlabel('x_1'); ylabel('x_2');
    end 
    set(gca, 'PlotBoxAspectRatio', [1,1,1]);
    hold off;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Slow Feature Analysis
%
% create a SFA object
hdl = sfa2_create(opts.pp_range, xp_dim(opts.pp_range), 'PCA2', 'ORD1', 0, opts);
% perform the preprocessing step
sfa_step(hdl, x, 'preprocessing');
% perform the expansion step 
% IMPORTANT: add patterns one class at a time (!)
for n=1:nclass
    sfa_step(hdl, x(realclass==opts.classes(n),:), 'expansion','CLASSIF');
end
% close the algorithm
sfa_step(hdl, x, 'sfa','SVD_SFA');        
%sfa_step(hdl, x, 'sfa','GEN_EIG');        
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% compute the SFA output signal on the training data
y = sfa_execute(hdl, x);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% train the Gaussian classifier using training data and predict with it
%
out = y(:,1:opts.gaussdim);
hdg = gaussCreate(nclass,out);
predC = gaussClassifier(hdg,out,realclass,'train',opts.ALIGNED);
%
% how good/bad were another Gaussian classifier based directly on the input
% 'x' (bypassing SFA)?
hdx = gaussCreate(nclass,x);
predX = gaussClassifier(hdx,x,realclass,'train',opts.ALIGNED);

errors = numel(find(predC~=realclass));    
errorX = numel(find(predX~=realclass));    
errtrn = [errors errorX]/numel(realclass);
fprintf('                             with SFA   w/o SFA\n');    
fprintf('Error rate on training set = %g%%       %g%%\n', errtrn*100);    

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if size(xtst,1)>0
    % if there are test data:
    % compute the SFA output signal on the test data and predict with it
    ytst = sfa_execute(hdl, xtst);
    outtst = ytst(:,1:opts.gaussdim);
    predT = gaussClassifier(hdg,outtst,-1,'apply');
    predX = gaussClassifier(hdx,xtst,-1,'apply');
    errors = numel(find(predT~=realc_tst));    
    errorX = numel(find(predX~=realc_tst));    
    errtst = [errors errorX]/numel(realc_tst);
    fprintf('Error rate on    test set  = %g%%       %g%%\n', errtst*100);    
else
    errtst = [];
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Some more plots
%
if opts.dographics>=1
    % plot the output of the 3 slowest varying functions on training data
    [rc_sort,idx]=sort(realclass);     % sort the vowel data by increasing class number
    figure(1);
    for i=1:3
        subplot(2,3,i+spl); plot(1:trnnum, y(idx,i));
        set(gca, 'PlotBoxAspectRatio', [1,1,1]);
        title(sprintf('output of slow varying function y_%1d(t)',i));
        xlabel('index'); ylabel(sprintf('y_%1d(t)',i));
    end
    spl=spl+3;

    % scatter plot output signal (first 2 dims, all classes in different
    % colors)
    figure(1); 
    spl=spl+1; subplot(2,3,spl);
    hold on; 
    for n=1:nclass
        ind = find(realclass==opts.classes(n));
        plot(y(ind,1),y(ind,2),'Marker','.','MarkerSize',5,'Color',colors(n,:),'LineStyle','none');
        title('the first two output components');
        xlabel('y_1'); ylabel('y_2');
    end 
    set(gca, 'PlotBoxAspectRatio', [1,1,1]); 
    hold off;

    % plot the final class predictor (on training data)
    spl=spl+1; subplot(2,3,spl); 
    hold on;
        predC_sort = predC(idx);
        plot(1:trnnum, predC_sort);
        for n=1:nclass
            ind = find(rc_sort==n);
            plot(ind,predC_sort(ind),'Color',colors(n,:),'LineStyle','-');
        end
        hold off;
        set(gca, 'PlotBoxAspectRatio', [1,1,1]);
        title(sprintf('class predicted by SFA'));
        xlabel('index'); ylabel('predC');
        ylim([0 nclass+1]);
    %print -f1 -dpng -r300 images/class_demo.png
end % dographics>=1

if opts.dographics>=3
    figure(2); clf;
    r=2; c=2;
    for i=1:min([opts.nclass-1 r*c])
        subplot(r,c,i);
        plot(1:trnnum, y(idx,i));
        set(gca, 'PlotBoxAspectRatio', [1,1,1]);
        title(sprintf('output of slow varying function y_%1d(t)',i));
        xlabel('index'); ylabel(sprintf('y_%1d(t)',i));        
    end
end % dographics>=3

res.errtrn=errtrn;
res.errtst=errtst;
res.y=y;
res.ytst=ytst;
res.predT=predT;        % predictions on test set with SFA + GaussClassifier
res.predX=predX;        % predictions on test set w/o SFA (only GaussClassifier)
