% This demo reproduces the driving force example of the logistic map (Figure 2)
% from Wiskott, L. (2003), "Estimating Driving Forces of Nonstationary Time
% Series with Slow Feature Analysis",
% http://arxiv.org/abs/cond-mat/0312317. 
%
% Here the driving force gam(t) consist of a slow part gamS(t) and a fast
% part gamF(t). See Konen, W. (2009), "How slow is slow? SFA detects signals
% that are slower than the driving force"
% http://arxiv.org/abs/cond-mat/????. 
%  
% INPUT:
%       m               embedding dimension (def. 19)
%       tau             delay (def. 1)
%       nuf             base frequency, the larger the faster varies the 
%                       driving force (def. 50)
%       q               'chaos parameter': for q<0.3 the logistic map is
%                       fully in the chaotic regime, for 0.3<q<3.9 there
%                       are more and more non-chaotic parts in the time
%                       series (def. 0.1)
%       dographics      (def. 2) if >= 2, make some plots
%       noiseperc       (def. 0) e.g. 0.05 means 'add 5% noise to the data w
%       method          (def. 'SVD_SFA') either 'SVD_SFA', the new, numeric stable  
%                       method, or 'GEN_EIG' (the old version of [Berkes03])
% OUTPUT: Struct res with tags:
%       cor             cor(1): correlation between gam(t) and y_1(t)
%                       cor(2): correlation between gamS(t) and y_1(t)
%       slowness        slowness for gam(t), a*y_1(t)+b, original w(t),gamS(t)
%       eta             alternative slowness
%
% EXAMPLE:
%       drive1(19,1,50,1.9,2,0,method)
% This will result in a wrong slow signal for method = 'GEN_EIG', but to a
% well aligned slow signal for method = 'SVD_SFA'.
%
% Author: Wolfgang Konen, May 2009 - Nov 2009
%
function res = drive1(m,tau,nuf,q,dographics,noiseperc,method) 

    dbg = 0;
    pp_type = 'PCA';        % preprocessing type
    
    if nargin<7, method='SVD_SFA'; end;     % the recommended way
    if nargin<6, noiseperc=0.00; end;
    if nargin<5, dographics=2; end;
    if nargin<4, q=0.1; end;
    if nargin<3, nuf=50; end;
    if nargin<2, tau=1; end;
    if nargin<1, m=19; end;

    % create the input signal
    T0 = 5990;
    T = T0+m*tau;
    t0 = linspace(0, T0-1, T0);
    t = linspace(0, T-1, T);

                    % REMARKABLE: If we set nuf=50, then the driving force is
                    % no longer very 'slow'. However, it contains a slower
                    % subcomponent. And - bingo - SFA detects the slower 
                    % subcomponent and thus a signal which is more slowly 
                    % than the driving force itself (has a smaller eta 
                    % than gam(t), see below).
                    % [nuf=50 requires however m=20 or higher to reconstruct 
                    % the slow component with good accuracy]
    %
    % Variant 1: driving force gam is composed of two sine waves at
    % frequency nuf*0.0047 and nuf*0.0005 (this is what we wrongly called 'beat'
    % above). SFA nicely learns to detect either one of these components,
    % depending on nuf
    %gam = sin(nuf*0.0021*t).*cos(nuf*0.0026*t);        % 0.0026
    gam = (sin(nuf*0.0047*t) + sin(nuf*0.0005*t))/2;    % equivalent to line above
    %gam = (0*sin(nuf*0.0047*t) + sin(nuf*0.0005*t))/2;     % activate this line for arxiv2009_SFA2.m
    %gam = (sin(nuf*0.0012*t) + sin(nuf*0.0005*t))/2;     
    gamS = + sin(nuf*0.0005*t);     % only the slow component
    gamF = + sin(nuf*0.0047*t);     % only the fast component
    %
    % Variant 2: driving force is composed of two nearby sine waves at
    % frequency nuf*0.0021 and nuf*0.0026. Now this contains a modulation (beat) at
    % frequency nuf*0.0005, but the modulation is hardly detected by SFA, no matter how
    % high we put nuf.
    %    gam = (sin(nuf*0.0021*t) + sin(nuf*0.0026*t))/2; 
    %    gamS = + cos(nuf*0.0005*t);     % only the slow component
    %    gamF = + sin(nuf*0.0047*t);     % only the fast component

    w = zeros(T,1)+0.1;
%     ntrans=30;
%     for i=1:ntrans
%         %avoid transient signal
%         w(i+1)=(4.0-q + gam(i)*0.1)*w(i)*(1-w(i));     % logistic map, Eq. (22)
%     end    
%     w(1:2)=w((ntrans-1):ntrans);
    for i=1:T-1
        w(i+1)=(4.0-q + gam(i)*0.1)*w(i)*(1-w(i));     % logistic map, Eq. (22)
    end;
    
    if dographics>=2
        % plot the input signal
        figure(1);
        pskip=1;
        subplot(3,1,1); plot(t(1:pskip:end),w(1:pskip:end),'.r','MarkerSize',4);
        % why '1:pskip:end'? - to make the PDF of the plot with pskip=2
        % somewhat smaller (360kB instead of 600kB). Default is pskip=1.
        title('input signal w(t)');
        ylabel('w(t)'); xlabel('t'); 
        set(gca, 'XLim', [0,6000]);
    end
    
    noise=randn(T,1)*noiseperc;
    fprintf('noise = %9.2g %%\n',mean(abs(noise(:)))/mean(abs(w(:)))*100);
    w=w+noise;

    x = zeros(T0,m);
    for i=1:m
        low = 1+tau*(i-1); 
        hig = T0+tau*(i-1);
        x(:,i)=w(low:hig);
    end


    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %
    % Slow Feature Analysis
    %
    opts.dbg=0
    [y, hdl] = sfa2(x,method,pp_type,opts);

    % plot the output of the slowest varying function
    y_1 = y(:,1);
    y_2 = y(:,2);
    
    if dographics>=2
        subplot(3,1,2); plot(t0(1:pskip:end), y_1(1:pskip:end),'.b','MarkerSize',4);
        % why '1:pskip:end'? - to make the PDF of the plot with pskip=2
        % somewhat smaller (360kB instead of 600kB). Default is pskip=1.
        title('output of the slowest varying function y_1(t)'); ylabel('y_1(t)');
        set(gca, 'YLim', [min(y_1),max(y_1)]*1.1);
        xlabel('t'); 
    end
    
    g_m = gam(1+fix(m/2):T0+fix(m/2))';     % why '+m/2'? If we pick up the driving
                                            % force here, there is no phase
                                            % shift to the slowest SFA signal
    g_S = gamS(1+fix(m/2):T0+fix(m/2))';
    g_F = gamF(1+fix(m/2):T0+fix(m/2))';

    % find the linear transformation [a b] which brings a*y(t)+b and g_m(t) in
    % optimal alignment: z = [a b] = X\g_m is the vector which solves the 
    % linear system X*z = g_m optimally in the LS sense.
    X = [y_1 ones(size(y_1))]; 
    z=X\g_m; a=z(1); b=z(2);
    z0=X\g_S; a0=z0(1); b0=z0(2);
    
    if dographics>=2
    % Now plot a*y(t)+b as blue points and overlay g_m(t) as green line:
    subplot(3,1,3); plot(t0, a*y_1+b,'.k',t0,g_m,'-g',...
                         t0, a0*y_1+b0,'.b',t0,g_S,'--r',...
                         'MarkerSize',4,'LineWidth',1.3);
    set(gca, 'YLim', [min(gam)-0.2,max(gam)+0.05]);
    title('true driving force \gamma(t) [green], its slow part \gamma_S(t) [red], alignment of y_1(t) [blue]');
    ylabel('\gamma(t)'); xlabel('t');
    text(5,-0.8,['\nu_f=' num2str(nuf)],'BackgroundColor',[1,0.9,0.6]);
    set(gca, 'XLim', [1 500]);          % activate this line for nuf=50 to see 
                                        % the beat (Schwebung)
       
    figure(1);
    dummy=1;    % set a breakpoint here if you want to see the figure 1 for every step
                % (with 'dbcont' in Command Window)
    
    if (~isempty(find([20 40 60]==nuf)))
        figure(4); 
        if (nuf==20)
            clf; set(gcf, 'Position', [189 177 866 540]);
            subplot(3,1,1);
        end
        if nuf==40, subplot(3,1,2);, end;
        if nuf==60, subplot(3,1,3);, end;
        plot(t0, a*y_1+b,'.b',t0,g_m,'-g',...
             t0, a0*y_1+b0,'.b',t0,g_S,'--r',...
                         'MarkerSize',4,'LineWidth',1.3);
        set(gca, 'YLim', [min(gam)-0.2,max(gam)+0.05]);
        title('true driving force \gamma(t) [green], its slow part \gamma_S(t) [red], alignment of y_1(t) [blue]');
        ylabel('\gamma(t)'); xlabel('t');
        text(5,-0.8,['\nu_f=' num2str(nuf)],'BackgroundColor',[1,0.9,0.6]);
        set(gca, 'XLim', [1 500]);
    end  %% ~isempty
    end  %% dographics

    % correlation and slowness as indicative numbers:
    % (slowness = mean of |finite differences| )
    corr=corrcoef(g_m,y_1); cor(1)=corr(1,2);
    corr=corrcoef(g_S,y_1); cor(2)=corr(1,2);
    if dographics>=3
        % Plot |correlation| and slowness for *all* SFA output signals,
        % ordered by their eigenvalue size (Where are sudden jumps?)
        C = abs(corrcoef([g_S,y])); 
        figure(4), subplot(2,1,1); plot(C(1,2:end));
        for i=1:size(y,2)
            etay(i)=etaval(y(:,i),T0);
        end
        subplot(2,1,2); plot(etay);
    end
    if isnan(cor(1))        % should not happen (but can, if nuf=0)
        warning('correlation with g_m contains NaN');
    end
    if isnan(cor(2))        % should not happen (but can, if nuf=0)
        warning('correlation with g_S contains NaN');
    end
    slowness(1) = mean(abs(diff(g_m)));
    slowness(2) = mean(abs(diff(a*y_1+b)));
    slowness(3) = mean(abs(diff(w)));
    slowness(4) = mean(abs(diff(g_S)));
    eta(1) = etaval(g_m,T0); %sqrt(mean(diff(g_m).^2))/sqrt(mean(g_m.^2));
    eta(2) = etaval(a*y_1+b,T0); %sqrt(mean(diff(a*y_1+b).^2))/sqrt(mean((a*y_1+b).^2));
    eta(3) = etaval(w,T); %sqrt(mean(diff(w).^2))/sqrt(mean(w.^2));
    eta(4) = etaval(g_S,T0); %sqrt(mean(diff(g_S).^2))/sqrt(mean(g_S.^2));
    
    res.cor=cor;
    res.eta=eta;
    res.slowness=slowness;
    res.g_m=g_m; 
    res.g_S=g_S;
    res.t0=t0;
    res.y_1=y_1;
    res.hdl=hdl;
end 

% TODO: 
%   - why is the logistic map task easier than the tent map task?
%   - with nuf=20, m=20 and 'XLim', [1 500] we see a phase shift of m/2 between y_1 and
%   gam. Can it be corrected? 