% Function drive1_sweep calls drive1.m for different values of nuf
% (the fastness of the driving force) and collects the results in matrices.
% INPUT
%       m               embedding dimension
%       q               'chaos parameter': for q<0.33 the logistic map is
%                       fully in the chaotic regime, for 0.33<q<0.53 a mix
%                       of chaotic & non-chaotic parts, for 0.53<q<3.9  
%                       fully non-chaotic time series
%       fmin,fmax,fstep
% RESULTS
%       nufmat          [fmin : fstep : fmax]
%       corrmat(:,4)    correlation with gam(t), gam0(t)
%       slowmat(:,4)    slowness for gam(t), y_1(t), original w(t), gam0(t)
%       etamat(:,4)     eta for gam(t), y_1(t), original w(t), gam0(t)
%       ptrans          phase transition frequency
% This demo shows a phase transition as nu_f increases, because the slower
% component comes into reach.
%
% Author: Wolfgang Konen, May 2009 - Aug 2009

function [ptrans, corrmat] = drive1_sweep(m,q,fmin,fmax,fstep,dographics)
    global MOV_STRUCTS 
    
    sfa_clearall;
    
    if ~exist('q','var'),   q = 0.9;, end;              % 'chaos factor'
    if ~exist('m','var'),   m = 19;, end;               % embedding dimension
    if ~exist('fmax','var'),    fmax = 80;, end; 
    if ~exist('fstep','var'),   fstep = 4;, end; 
    if ~exist('fmin','var'),    fmin = fstep;, end; 
    if ~exist('dographics','var'),  dographics = 2;, end; 
    
    if dographics>=2
        % init graphics
        figure(1); clf; set(gcf, 'Position', [89 77 866 540]);
    end 

    tau=1;                      % delay
    
    
    nufmat = [fmin : fstep : fmax];
    t = nufmat;
    corrmat = zeros(numel(t),2);
    slowmat = zeros(numel(t),4);
    etamat = zeros(numel(t),4);
    for n=1:numel(t)
        nuf = nufmat(n);
        MOV_STRUCTS.n = n;
        
        res = drive1(m,tau,nuf,q,dographics);
        
        if dographics>=2
            MOV_STRUCTS.mov3(n) = getframe(gcf);    % getframe(gcf) gets the whole window
            % Side effect: getframe(gcf) ensures that the actual graphics
            % window is updated and displayed while the MATLAB code is running.
            % If you do not want that figure windows pop up all the time,
            % disable this line
        end
        
        corrmat(n,:) = res.cor;
        slowmat(n,:) = res.slowness;
        etamat(n,:) = res.eta;
        
        fprintf('correlation coeff with gamma: %7.4f,     with gamma_0: %7.4f\n',res.cor);
        fprintf('slowness of gam, a*y_1+b, w and gam0: %8.5f,%8.5f,%8.5f,%8.5f\n', res.slowness);
        fprintf('eta-value of gam, a*y_1+b, w and gam0: %8.5f,%8.5f,%8.5f,%8.5f\n', res.eta);
    end
    
    % Calculation of phase transition frequency: crosspoint of correlation
    % curves
    ptrans = min(nufmat(find(abs(corrmat(:,2))>abs(corrmat(:,1)))));
    if isempty(ptrans)
        fprintf('No phase transition observed\n');
        ptrans = NaN; cptrans=NaN;
    else
        if ptrans==min(nufmat)
            warning('ptrans detected at lower bound of nu! Consider lowering fmin!');
            cptrans=corrmat(1,1);   % dummy for plot
        else
            % localize ptrans better by linear approximation (intersection
            % of correlation curves)
            ind = find(nufmat==ptrans);
            a = abs(corrmat(ind-1,2)); b = abs(corrmat(ind,2));
            c = abs(corrmat(ind-1,1)); d = abs(corrmat(ind,1));
            s = (c-a)/((b-a)+(c-d));
            ptrans = nufmat(ind-1) + s*fstep;
            cptrans = a + s*(b-a);
        end
        fprintf('Phase transition at nu_f=%7.3f\n',ptrans);
    end 
    
    if dographics>=1
        %Xmat = slowmat;
        Xmat = etamat;
        figure(2); clf; axes('FontSize',16,'Box','on');  
        plot(t,abs(corrmat(:,1)),'-.b', ...
             t,abs(corrmat(:,2)),'--k',  ...
             t,Xmat(:,2)./Xmat(:,1),'-r','LineWidth',2);
             %t,30*Xmat(:,2),'-k', ...
        legend('corr with \gamma','corr with \gamma_S','\eta_{SFA}/\eta_{\gamma}', ...
                'Location','East'); 
                %'30S_{SFA}', ...
        hold on;
        plot(ptrans,cptrans,'.k','MarkerSize',28);
        title(['Phase transition at \nu_f=' num2str(ptrans,'%5.1f')],'FontSize',16); 
        xlabel('\nu_f','FontSize',16);
        text(fmax/12,0.1,['q=' num2str(q)],'BackgroundColor',[1,0.9,0.6],'FontSize',16);
        set(gca, 'YLim', [0 1.05]);
        hold off;
        
        getframe(gcf);    % show figure(2) while MATLAB-code is running
        
    %     figure(3);
    %     plot(t,Xmat(:,1),'-.b',t,Xmat(:,2),'-r','LineWidth',2);
    %     legend('\gamma(t)','y_1(t)');
    %     title('Variance of time derivative for driving force \gamma(t) and slowest SFA signal y_1(t)');
    %     xlabel('\nu_f');

        %print -f1 -depsc2 -r300 images/logmap-drive-m11-nuf20.eps   % 'epsc2 important for color (!)
        %print -f4 -depsc2 -r300 images/logmap-drive-m19-nuf20-40-60.eps
        %print -f2 -depsc2 -r300 images/logmap-phasetrans-m19-q04.eps
        %print -f1 -dpng -r300 images/logmap-drive-beat-nuf50.png
        %print -f2 -dpng -r200 images/logmap-phase-transition.png
        %print -f3 -dpng -r200 images/logmap-slowness.png
    end
    
    if dographics>=2
        %fps=10;
        %figure(1);
        %movie(MOV_STRUCTS.mov3,2,fps);
        % to save the movie on file:
        %movie2avi(MOV_STRUCTS.mov3, 'test.avi','fps',10);
    end
    
    dummy=1;            % just to set a breakpoint

end

