% This demo reproduces the driving force example of the logistic map (Figure 2)
% from Wiskott, L. (2003), "Estimating Driving Forces of Nonstationary Time
% Series with Slow Feature Analysis", 
% http://arxiv.org/abs/cond-mat/0312317 
%
% Author: Wolfgang Konen, May 2009

msg = ['This demo reproduces an example from Wiskott, L. (2003),' ...
       ' "Estimating Driving Forces of Nonstationary Time' ...
       ' Series with Slow Feature Analysis",' ...
       ' http://arxiv.org/abs/cond-mat/0312317 \n\n'];

%fprintf(msg);


% init graphics
figure(1); clf; set(gcf, 'Position', [89 77 866 538]);

% create the input signal
m = 27;                     % embedding dimension
tau=1;                      % delay
T0 = 5990;
T = T0+m*tau;
t0 = linspace(0, T0-1, T0);
t = linspace(0, T-1, T);

                % REMARKABLE: If we set nu_f=20, then the driving force is
                % no longer very 'slow'. However, it contains a beat
                % (Schwebung), which is much slower. And - bingo - SFA
                % detects the slower beat and thus a signal which is more
                % slowly than the driving force itself (smaller 'slowness' 
                % than gam(t), see below).
nu_f = 20;                    
gam = -sin(nu_f*0.0021*t).*cos(nu_f*0.0026*t);
gam = gam*0.5-0.54;         % Experiment 4: put the driving force in the 
                            % non-chaotic range of the logistic map
 
w = zeros(T,1)+0.1;
for i=1:T-1
    w(i+1)=(3.6+0.4*gam(i))*w(i)*(1-w(i));      % logistic map, Eq. (19)
end;
%w = w+randn(size(w))*2e-08; % we don't need noise injection in Exp. 4 (!)

% plot the input signal
subplot(3,1,1); plot(t,w,'.r','MarkerSize',4);
title('input signal w(t)');
ylabel('w(t)');
set(gca, 'XLim', [1 6000]);
set(gca, 'YLim', [0.3 1]);

x = zeros(T0,m);
for i=1:m
    low = 1+tau*(i-1); 
    hig = T0+tau*(i-1);
    x(:,i)=w(low:hig);
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Slow Feature Analysis
%
% create a SFA object
opts.dbg=0;                                      
hdl = sfa2_create(m, xp_dim(m), 'PCA', 'ORD1', 0, opts);
% perform the preprocessing step
sfa_step(hdl, x, 'preprocessing');
% perform the expansion step
sfa_step(hdl, x, 'expansion');
% close the algorithm

                                         % Experiment 4:  
sfa_step(hdl, x, 'sfa', 'GEN_EIG');      % activate this line for exp4-GEN*.png images        
%sfa_step(hdl, x, 'sfa', 'SVD_SFA');      % activate this line for exp4-SVD*.png images
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% compute the output signal
y = sfa_execute(hdl, x);

% it would have been quicker (but less instructive) to write:
% [y, hdl] = sfa2(x);

% plot the output of the slowest varying function
y_1 = y(:,1);
subplot(3,1,2); plot(t0, y_1,'.b','MarkerSize',4);
set(gca, 'YLim', [min(y_1),max(y_1)]*1.1);
title('output of the slowest varying function y_1(t)');
xlabel('t'); ylabel('y_1(t)');

g0 = fix(m/2);              % why 'fix(m/2)'? If we pick up the driving
g_m = gam(1+g0:T0+g0)';     % force here, there is no phase
                            % shift to the slowest SFA signal            
                                        
% find the linear transformation [a b] which brings a*y(t)+b and g_m(t) in
% optimal alignment: z = [a b] = X\g_m is the vector which solves the 
% linear system X*z = g_m optimally in the LS sense.
X = [y_1 ones(size(y_1))]; 
z=X\g_m; a=z(1); b=z(2);
% Now plot a*y(t)+b as blue points and overlay g_m(t) as green line:
subplot(3,1,3); plot(t0, a*y_1+b,'.b',t0,g_m,'-g','MarkerSize',4,'LineWidth',1.3);
set(gca, 'YLim', [min(gam)-0.2,max(gam)+0.05]);
title('true driving force \gamma(t) [green], alignment of y_1(t) [blue]');
ylabel('\gamma(t)');
set(gca, 'XLim', [1 500]);          % activate this line for nu_f=50 to see 
                                    % the beat (Schwebung)
 
% correlation and slowness as indicative numbers:
% (slowness = mean of |finite differences| )
cor=corrcoef(g_m,y_1); cor=cor(1,2);
slowness(1) = mean(abs(diff(g_m)));
slowness(2) = mean(abs(diff(a*y_1+b)));
slowness(3) = mean(abs(diff(w)));
fprintf('correlation coeff: %7.4f,     mean(y_1^2): %12.3e\n',cor,mean(y_1.*y_1));
fprintf('slowness  of gam, a*y_1+b and w: %8.5f,%8.5f,%8.5f\n', slowness);
eta(1) = sqrt(mean(diff(g_m).^2));
eta(2) = sqrt(mean(diff(a*y_1+b).^2));
eta(3) = sqrt(mean(diff(w).^2));
fprintf('eta-value of gam, a*y_1+b and w: %8.5f,%8.5f,%8.5f\n', eta);

Cslow = (y'*y)/numel(y_1);
%disp('Covariance matrix of the slow signals:');
%disp(Cslow);

%print -f1 -dpng -r200 images/exp4-GEN-m06-nuf40.png
