% Function drive_sweep calls drive1 for different values of fac (the fastness 
% of the driving force and collects the results in matrices 
%       corrmat(:,1)    correlation
%       slowmat(:,3)    slowness for gam(t), y_1(t), original w(t)
% This demo shows a phase transition when fac increases, because the beat
% (Schwebung) comes into reach.
%
% Author: Wolfgang Konen, May 2009

function drive4_sweep()
    % init graphics
    figure(1); clf; set(gcf, 'Position', [89 77 866 538]);

    m = 15;
    fmax = 20;
    fstep = 1;
    %method = 'GEN_EIG';
    method = 'SVD_SFA';
    
    t = linspace(fstep,fmax,fmax/fstep);
    corrmat = zeros(fmax/fstep,1);
    slowmat = zeros(fmax/fstep,3);
    for fac=fstep:fstep:fmax
        
        [cor,slowness] = drive4(m,fac,method);
        
        corrmat(fac/fstep) = cor;
        slowmat(fac/fstep,:) = slowness;
        
        fprintf('correlation coeff: %7.4f\n',cor);
        fprintf('slowness of gam, a*y_1+b and w: %8.5f,%8.5f,%8.5f\n', slowness);
    end
    figure(2); 
    plot(t,abs(corrmat),'-.b', ...
         t,slowmat(:,2)./slowmat(:,1),'-r', ...
         t,30*slowmat(:,2),'-k','LineWidth',2);
    legend('correlation','S_{SFA}/S_{\gamma}','30S_{SFA}'); 
    title('Phase transition at \nu=7'); xlabel('\nu');
    set(gca, 'YLim', [0 1.2]);
    figure(3);
    plot(t,slowmat(:,1),'-.b',t,slowmat(:,2),'-r','LineWidth',2);
    legend('\gamma(t)','y_1(t)');
    title('Variance of time derivative for driving force \gamma(t) and slowest SFA signal y_1(t)');
    xlabel('\nu');
    
    print -f1 -dpng -r300 images/log2-drive-beat-nuf20.png
    print -f2 -dpng -r200 images/log2-phase-transition.png
    print -f3 -dpng -r200 images/log2-slowness.png
    
    dummy=1;            % just to set a breakpoint

end

% This demo reproduces the driving force example of the logistic map (Figure 2)
% from Wiskott, L. (2003), "Estimating Driving Forces of Nonstationary Time
% Series with Slow Feature Analysis", 
% http://arxiv.org/abs/cond-mat/0312317 
% INPUT:
%       fac             factor, the larger the faster varies the driving
%                       force 
% OUTPUT:
%       cor             correlation between gam(t) and y_1(t)
%       slowmat(:,3)    slowness for gam(t), a*y_1(t)+b, original w(t)
%
function [cor,slowness] = drive4(m,nu_f,method) 
    % create the input signal
    %m =  20;                     % embedding dimension
    tau=1;                      % delay
    T0 = 5990;
    T = T0+m*tau;
    t0 = linspace(0, T0-1, T0);
    t = linspace(0, T-1, T);

                    % REMARKABLE: If we set nu_f=50, then the driving force is
                    % no longer very 'slow'. However, it contains a beat
                    % (Schwebung), which is much slower. And - bingo - SFA
                    % detects the slower beat and thus a signal which is more
                    % slowly than the driving force itself (smaller 'slowness' 
                    % than gam(t), see below).
    %nu_f = 50;                    
    gam = -sin(nu_f*0.0021*t).*cos(nu_f*0.0026*t);
    gam = gam*0.5-0.54;         % Experiment 4: put the driving force in the 
                                % non-chaotic range of the logistic map
 
    w = zeros(T,1)+0.1;
    for i=1:T-1
        w(i+1)=(3.6+0.4*gam(i))*w(i)*(1-w(i));      % logistic map, Eq. (19)
    end;
    %w = w+randn(size(w))*2e-08; % we don't need noise injection in Exp. 4 (!)

    % plot the input signal
    subplot(3,1,1); plot(t,w,'.r','MarkerSize',4);
    title('input signal w(t)');
    ylabel('w(t)');
    set(gca, 'XLim', [1 6000]);
    set(gca, 'YLim', [0.3 1]);

    x = zeros(T0,m);
    for i=1:m
        low = 1+tau*(i-1); 
        hig = T0+tau*(i-1);
        x(:,i)=w(low:hig);
    end

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %
    % Slow Feature Analysis
    %
    % create a SFA object
    opts.dbg=0;                                      
    hdl = sfa2_create(m, m,'PCA', 'ORD1', 0, opts); 
    %hdl = sfa2_create(m, xp_dim(m),'PCA', 'ORD1', 0, opts);
    % perform the preprocessing step
    sfa_step(hdl, x, 'preprocessing');
    % perform the expansion step
    sfa_step(hdl, x, 'expansion');
    % close the algorithm

    sfa_step(hdl, x, 'sfa', method);      
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    % compute the output signal
    y = sfa_execute(hdl, x);

    % plot the output of the slowest varying function
    y_1 = y(:,1);
    subplot(3,1,2); plot(t0, y_1,'.b','MarkerSize',4);
    set(gca, 'YLim', [min(y_1),max(y_1)]*1.1);
    title('output of the slowest varying function y_1(t)');
    xlabel('t'); ylabel('y_1(t)');

    g_m = gam(1+fix(m/2):T0+fix(m/2))';     % why '+m/2'? If we pick up the driving
                                            % force here, there is no phase
                                            % shift to the slowest SFA signal

    % find the linear transformation [a b] which brings a*y(t)+b and g_m(t) in
    % optimal alignment: z = [a b] = X\g_m is the vector which solves the 
    % linear system X*z = g_m optimally in the LS sense.
    X = [y_1 ones(size(y_1))]; 
    z=X\g_m; a=z(1); b=z(2);
    % Now plot a*y(t)+b as green points and overlay gam(t) as blue line:
    subplot(3,1,3); plot(t0, a*y_1+b,'.b',t0,g_m,'-g','MarkerSize',4,'LineWidth',1.3);
    set(gca, 'YLim', [min(gam)-0.2,max(gam)+0.05]);
    title('true driving force \gamma(t) [green], alignment of y_1(t) [blue]');
    ylabel('\gamma(t)');
    set(gca, 'XLim', [1 500]);          % activate this line for fac=50 to see 
                                        % the beat (Schwebung)

    % correlation and slowness as indicative numbers:
    % (slowness = mean of |finite differences| )
    cor=corrcoef(g_m,y_1); cor=cor(1,2);
    slowness(1) = mean(abs(diff(g_m)));
    slowness(2) = mean(abs(diff(a*y_1+b)));
    slowness(3) = mean(abs(diff(w)));
    
end 

% TODO: 
%   - why is the logistic map task easier than the tent map task?
%   - with fac=20, m=20 and 'XLim', [1 500] we see a phase shift of m/2 between y_1 and
%   gam. Can it be corrected? 