function [predC,prob] = gaussClassifier(hdl,y,realC,method,aligned)
% gaussClassifier: train or apply a Gaussian classifier HDL, which has
% been created before with gaussCreate.
% INPUT:
%   y       K x M matrix where K is the total number of patterns and M is the
%           number of variables used for classification. I.e. each row of
%           y contains the data for one pattern.
%   realC   1 x K matrix with NCLASS distinct real class labels needed only
%           for method='train'. In case of method='apply' realC is not 
%           used and can have any value
%   method  either 'train' (def.) or 'apply'
%   aligned =0: do not align the Gaussian classifiers with axes, use full
%           covariance matrix
%           =1 (def.): set the off-diagonals in covariance matrix to 0,
%           i.e. the Gaussian classifier is forced to be aligned with the
%           axes. This is more robust in the case where the data deviate
%           largely from a multivariate normal distribution. 
% OUTPUT: 
%   predC   1 x K matrix: the predicted class = argmax(prob(k,:))
%   prob    K x NCLASS matrix: prob(k,n) is the estimated probability that
%           pattern k belongs to class m
%
% Author: Wolfgang Konen, Cologne Univ. of Applied Sciences, Aug 2009 - Oct 2009
%
% See also: GAUSSCREATE, GAUSS_CLEARALL


    global GAUSS_STRUCTS    
    nclass = GAUSS_STRUCTS{hdl}.nclass; 
    if ~exist('method','var'), method='train';, end;
    if ~exist('aligned','var'), aligned=1;, end;
    
    if strcmp(method,'train')
        GAUSS_STRUCTS{hdl}.uni = unique(realC);        % the unique labels 
        % (during training we must have at least one record from each
        % class, during 'apply' there can be fewer class labels present, 
        % therefore we store GAUSS_STRUCTS{hdl}.uni for later reference)
        for n=1:nclass
            ind = find(realC==GAUSS_STRUCTS{hdl}.uni(n));
            GAUSS_STRUCTS{hdl}.X0(n,:) = mean(y(ind,:));
            arg = y(ind,:)-repmat(GAUSS_STRUCTS{hdl}.X0(n,:),numel(ind),1);
            COV = arg' * arg/(numel(ind)-1);
            if aligned
                % zero all off-diagonal elements >> ensures that the densitiy
                % function has ellipsoidal equipotential surfaces and that
                % the ellipsoid axes are *aligned* with the coordinate axes
                COV = diag(diag(COV));
            end
            GAUSS_STRUCTS{hdl}.COV(:,:,n) = COV;
            GAUSS_STRUCTS{hdl}.iCOV(:,:,n) = inv(GAUSS_STRUCTS{hdl}.COV(:,:,n));
            GAUSS_STRUCTS{hdl}.f0(n) = 1/(2*pi)^(GAUSS_STRUCTS{hdl}.odim/2) ...
                                        /sqrt(det(GAUSS_STRUCTS{hdl}.COV(:,:,n)));
            GAUSS_STRUCTS{hdl}.P_c(n) = numel(ind)/numel(realC);
            
            DBG=0;
            if DBG>0 
                [X1,X2] = meshgrid(-2:.2:2,-2:.2:2);
                Z = 0*X1;
                if GAUSS_STRUCTS{hdl}.odim>2
                    X3 = GAUSS_STRUCTS{hdl}.X0(n,3:end);
                else 
                    X3 = [];
                end
                for i=1:size(X1,1)
                    for j=1:size(X1,2)
                        xvec=[X1(i,j) X2(i,j) X3] - GAUSS_STRUCTS{hdl}.X0(n,:); 
                        Z(i,j) = exp(-xvec*GAUSS_STRUCTS{hdl}.iCOV(:,:,n)*xvec'/2);                      
                    end
                end
                Z = Z*GAUSS_STRUCTS{hdl}.f0(n);
                fprintf('sum(Z)=%g\n',sum(Z(:)));
                figure(2); clf; 
                hold on;
                contour(X1,X2,Z);
                plot(y(ind,1),y(ind,2),'.r');
                hold off;
                xlabel('y_1'); ylabel('y_2');
                xlim([-2 2]); ylim([-2 2]);
                set(gca, 'PlotBoxAspectRatio', [1,1,1]);
            end % DBG>0
        end
        GAUSS_STRUCTS{hdl}.isTrained=1;
    elseif strcmp(method,'apply')
    else
        error(['method = ' method ' is not allowed']);
    end
    
    if GAUSS_STRUCTS{hdl}.isTrained==0
        error(sprintf('Gaussian Classifier hdl=%d is not trained >> cannot apply it',hdl));
    end
    % 
    % this is done for both methods, 'train' and 'apply':
    %
    prob = zeros(size(y,1),nclass);
    for k=1:size(y,1)
        sum_Py=0;
        for n=1:nclass
            arg = y(k,:)-GAUSS_STRUCTS{hdl}.X0(n,:);
            Py_c(n) = GAUSS_STRUCTS{hdl}.f0(n) * exp(-arg*GAUSS_STRUCTS{hdl}.iCOV(:,:,n)*arg'/2);
            sum_Py = sum_Py+Py_c(n)*GAUSS_STRUCTS{hdl}.P_c(n);
        end
        if sum_Py==0
            % the pattern y(k,:) is so far away from all Gaussians that we
            % can't tell the class >> random guess
            Pc_y = rand(1,nclass);
            warning(['sum_Py==0, pattern far outside for k=' num2str(k)]);
        else
            Pc_y = Py_c .* GAUSS_STRUCTS{hdl}.P_c' / sum_Py;
        end
        predC(k) = GAUSS_STRUCTS{hdl}.uni(find(Pc_y==max(Pc_y(:))));
        prob(k,:) = Pc_y;
    end
end