function sfa2_step(hdl, arg, step, method)
% SFA2_STEP a step in the SFA2 algorithm.
%   SFA2_STEP(HDL, ARG, STEPNAME, METHOD, DBG)
%   Do not use this function directly, use SFA_STEP instead.
%
%   Extension WK (Wolfgang Konen) 09/2009: new params, only relevant for steps 
%   'expansion' and 'sfa':
%       method  if step='sfa': method ='SVD_SFA' (def.) or ='GEN_EIG'
%               if step='expansion' or undefined: 
%                   method ='TIMESERIES' (def.) or ='CLASSIF'
%   Some parameters are taken from SFA_STRUCTS{hdl}.opts (see SFA_CREATE),
%   if it is a struct, otherwise defaults are set here: 
%       dbg     0 (def.) or >0, only relevant for step 'sfa', then SFA2_DBG is called
%       epsC    cutoff C-eigenvalues smaller than epsC*lammax (def. 1e-7)
%
%   See also SFA_STEP, SFA_CREATE
  
  global SFA_STRUCTS
  global LCOV_STRUCTS
  
  if isfield(SFA_STRUCTS{hdl}.opts,'dbg'), dbg=SFA_STRUCTS{hdl}.opts.dbg;
  else, dbg=0; end
  if isfield(SFA_STRUCTS{hdl}.opts,'epsC'), epsC=SFA_STRUCTS{hdl}.opts.epsC;
  else, epsC=1e-7; end
  
  %
  % things to do when switching from one step to another occurs
  %
  if nargin>2,
    old_step=SFA_STRUCTS{hdl}.step;
    %%% check step is init->preprocessing->expansion->sfa
    %%% clear step variables
    
    %%% init->preprocessing
    if strcmp(old_step,'init') & strcmp(step,'preprocessing'),
      fprintf('init preprocessing\n');
      if strncmpi(SFA_STRUCTS{hdl}.pp_type, 'PCA',3),   % either 'PCA','PCA2' or 'PCAVAR'
        SFA_STRUCTS{hdl}.lcov_hdl=lcov_create(size(arg,2));
      else % SFA1
        SFA_STRUCTS{hdl}.sfa1_hdl=sfa1_create(SFA_STRUCTS{hdl}.pp_range);
      end
      
    %%% preprocessing->expansion
    elseif strcmp(old_step,'preprocessing') & strcmp(step,'expansion'),
      % close preprocessing step
      fprintf('close preprocessing\n');
      
      if strcmpi(SFA_STRUCTS{hdl}.pp_type, 'SFA1'),
        sfa1_hdl=SFA_STRUCTS{hdl}.sfa1_hdl;
        sfa_step(sfa1_hdl, [], 'sfa');
        fprintf('(SFA1)\n');

        SFA_STRUCTS{hdl}.W0=SFA_STRUCTS{sfa1_hdl}.SF;
        SFA_STRUCTS{hdl}.D0=SFA_STRUCTS{sfa1_hdl}.DSF;
        SFA_STRUCTS{hdl}.avg0=SFA_STRUCTS{sfa1_hdl}.avg0;
        SFA_STRUCTS{hdl}.tlen0=SFA_STRUCTS{sfa1_hdl}.tlen0;

        %clean up
        sfa_clear(sfa1_hdl);
        SFA_STRUCTS{hdl}=rmfield(SFA_STRUCTS{hdl},'sfa1_hdl');

      else % 'PCA*'
        cvhdl=SFA_STRUCTS{hdl}.lcov_hdl;
        lcov_fix(cvhdl);
      
        if strcmpi(SFA_STRUCTS{hdl}.pp_type, 'PCA'),
            % the original preprocessing sphering by Berkes, using standard PCA
            %
            fprintf('whitening and dimensionality reduction (PCA)\n');
            [SFA_STRUCTS{hdl}.W0,SFA_STRUCTS{hdl}.DW0,SFA_STRUCTS{hdl}.D0]= ...
                lcov_pca(cvhdl,SFA_STRUCTS{hdl}.pp_range);
            SFA_STRUCTS{hdl}.avg0=LCOV_STRUCTS{cvhdl}.avg;
            SFA_STRUCTS{hdl}.tlen0=LCOV_STRUCTS{cvhdl}.tlen;

            % additional check: is covariance matrix of input data
            % ill-conditioned?   /WK/08/2009
            ev = eig(LCOV_STRUCTS{cvhdl}.COV_MTX);
            ev = sort(ev);
            cn = abs(ev(end)/ev(1));        % Condition number of B
            if cn > 1e10
                disp(['*** Warning ***: Covariance matrix of input data is ill-conditioned (CN = ' ...
                    num2str(cn,'%12.3e') ').']);
                disp('It is recommended to use SFA_STRUCTS{hdl}.pp_type = PCA2.');
            end

        elseif strcmpi(SFA_STRUCTS{hdl}.pp_type, 'PCA2'),
            % the improved preprocessing sphering by Konen, using SVD.
            % Redundant dimensions with eigenvalue close to zero are detected
            % and the corresponding rows in W0 removed.
            %
            fprintf('whitening and dimensionality reduction (PCA2)\n');
            [SFA_STRUCTS{hdl}.W0,SFA_STRUCTS{hdl}.DW0,SFA_STRUCTS{hdl}.D0]= ...
                lcov_pca2(cvhdl,SFA_STRUCTS{hdl}.pp_range);
            % lcov_pca2 will null the rows of SFA_STRUCTS{hdl}.W0 with too
            % small eigenvalues. Here we reduce the rows of W0 and the numbers
            % pp_range  and xp_range accordingly:
            pp_range = numel(find(sum(SFA_STRUCTS{hdl}.W0')~=0));
            SFA_STRUCTS{hdl}.pp_range = pp_range;
            SFA_STRUCTS{hdl}.xp_range=xp_dim(pp_range);
            SFA_STRUCTS{hdl}.sfa_range=min([xp_dim(pp_range) SFA_STRUCTS{hdl}.sfa_range]);
            SFA_STRUCTS{hdl}.W0 = SFA_STRUCTS{hdl}.W0(1:pp_range,:);

            SFA_STRUCTS{hdl}.avg0=LCOV_STRUCTS{cvhdl}.avg;
            SFA_STRUCTS{hdl}.tlen0=LCOV_STRUCTS{cvhdl}.tlen;

        elseif strcmpi(SFA_STRUCTS{hdl}.pp_type, 'PCAVAR'),
            % another preprocessing as done in [Wiskott&Sejnowski'02] which
            % does not use PCA, but simply shifts and scales the input data to
            % have zero mean and unit variance
            %
            fprintf('unit variance w/o dimensionality reduction (PCAVAR)\n');
            varmat = diag(diag(LCOV_STRUCTS{cvhdl}.COV_MTX));
            SFA_STRUCTS{hdl}.W0 = varmat^(-0.5);
            SFA_STRUCTS{hdl}.avg0=LCOV_STRUCTS{cvhdl}.avg;
            SFA_STRUCTS{hdl}.tlen0=LCOV_STRUCTS{cvhdl}.tlen;
        end

        %clean up
        lcov_clear(cvhdl);
        SFA_STRUCTS{hdl}=rmfield(SFA_STRUCTS{hdl},'lcov_hdl');

      end

      % init expansion step
      fprintf('init expansion step\n');
      insize=SFA_STRUCTS{hdl}.pp_range;
      if length(insize)==2, insize=insize(2)-insize(1)+1; end
      xpsize=SFA_STRUCTS{hdl}.xp_range; %xp_dim(insize);
      SFA_STRUCTS{hdl}.xp_hdl=lcov_create(xpsize);
      SFA_STRUCTS{hdl}.diff_hdl=lcov_create(xpsize);

    %%% expansion->sfa
    elseif strcmp(old_step,'expansion') & strcmp(step,'sfa'),
      % close expansion step
      fprintf('close expansion step\n');
      global LCOV_STRUCTS
      
      % expanded data
      xphdl=SFA_STRUCTS{hdl}.xp_hdl;
      lcov_fix(xphdl)
      SFA_STRUCTS{hdl}.avg1=LCOV_STRUCTS{xphdl}.avg;
      SFA_STRUCTS{hdl}.tlen1=LCOV_STRUCTS{xphdl}.tlen;
      xpsize=SFA_STRUCTS{hdl}.xp_range;

      dfhdl=SFA_STRUCTS{hdl}.diff_hdl;
      lcov_fix(dfhdl);              % ?? was missing before ?? /WK/

      % perform sfa step
      fprintf('perform slow feature analysis\n');
            
      sfa_range=SFA_STRUCTS{hdl}.sfa_range;
      if length(sfa_range)==1, sfa_int=1:sfa_range;
      else sfa_int=sfa_range(1):sfa_range(2); end
      
      
      if strcmp(method,'GEN_EIG') | dbg>0
          % additional check: is covariance matrix of expanded data 
          % ill-conditioned?   /WK/08/2009
          ev = eig(LCOV_STRUCTS{xphdl}.COV_MTX);
          ev = sort(ev);
          cn = abs(ev(end)/ev(1));        % Condition number of B
          if cn > 1e10
              disp(['*** Warning ***: Covariance matrix of expanded data is ill-conditioned (CN = ' ...
                  num2str(cn,'%12.3e') ').']);
              disp('It is recommended to use method="SVD_SFA" in SFA-step.');
          end
            
          %flag = 'qz';
          %flag = 'chol';
          [SFA_STRUCTS{hdl}.SF,D]= ...
            eig(LCOV_STRUCTS{dfhdl}.COV_MTX,LCOV_STRUCTS{xphdl}.COV_MTX); %,flag);
      end
      
      if strcmp(method,'SVD_SFA')
          % extension /WK/08/2009: first sphere expanded data with
          % LCOV_PCA2, taking care of zero or very small eigenvalues in B
          % by using the SVD approach
          % 
          disp('Using alternate [WisSej02] approach for SFA-calculation ...');
          [S, DS, BD]= lcov_pca2(xphdl);   % S: sphering matrix for expanded data (xphdl)
                                           % DS: de-sphering matrix, BD: eigenvalues 
                                           % of B (covariance matrix of expanded data)
          C = S*LCOV_STRUCTS{dfhdl}.COV_MTX*S';
          [W1,D1] = eig(C);
          SF1 = S'*W1;
%[tmp,D2,PC2]=svd(C);
% svd sorts the eigenvalues by itself (in *decreasing* order)
          if dbg>0 
             arg2=(arg-repmat(SFA_STRUCTS{hdl}.avg0, size(arg,1), 1))*SFA_STRUCTS{hdl}.W0';
             arg2=expansion(hdl, arg2);  
             arg2=arg2-repmat(SFA_STRUCTS{hdl}.avg1,size(arg2,1),1);
             sfa2_dbg(hdl,arg2,D,S,DS,BD,W1,D1) 
          end;
          
          SFA_STRUCTS{hdl}.SF = SF1;
          SFA_STRUCTS{hdl}.BD = BD;
          D=D1;
      end % 'method'
      
      % always calculate rank(B) (for diagnostics only)
      B = LCOV_STRUCTS{xphdl}.COV_MTX;
      rankB = rank(B);
      fprintf('rank(B)=%d\n',rankB);
      SFA_STRUCTS{hdl}.rankB = rankB;
      
      D=diag(D);
      lammax=max(D(:));
      fprintf('epsC*lammax=%e\n',epsC*lammax);
      [tmp,idx]=sort(D);
      if strcmp(method,'SVD_SFA')
          rankC = rank(C);
          fprintf('rank(C)=%d\n',rankC);
          %idx = idx(find(D(idx)~=0)); % 'SVD_SFA': exclude eigenvalues which
                                      % are *exactly* zero from further
                                      % analysis, since they correspond to
                                      % zeros in the sphering matrix
                                      % (degenerate dimensions)
          idx = idx(find(abs(D(idx))>epsC*lammax)); 
          disp(['min(D)= ' num2str(min(D))]);
          sfa_int = 1:numel(idx);
          SFA_STRUCTS{hdl}.sfa_range = numel(idx);
          SFA_STRUCTS{hdl}.myS=S;
        % REMARK: These statement were also beneficial for 'GEN_EIG', because
        % there it may happen in the degenerate case that some eigenvalues of
        % D become negative (??, and the corresponding eigenvectors contain
        % only noisy signals). However, we do not apply it here, because it
        % lets 'GEN_EIG' deviate from the original [Berkes03] code. And the
        % slow signals would still have the wrong variance.
      end
      SFA_STRUCTS{hdl}.DSF=D(idx(sfa_int));
      SFA_STRUCTS{hdl}.SF=SFA_STRUCTS{hdl}.SF(:,idx(sfa_int))';

      lcov_clear(xphdl);
      SFA_STRUCTS{hdl}=rmfield(SFA_STRUCTS{hdl},'xp_hdl');
      lcov_clear(dfhdl);
      SFA_STRUCTS{hdl}=rmfield(SFA_STRUCTS{hdl},'diff_hdl');

      fprintf('SFA2 closed\n');
    elseif ~strcmp(step,old_step),      % do nothing if step==old_step
      warning 'unknown step sequence'
      return
    end
      
    SFA_STRUCTS{hdl}.step=step;
  end % if (nargin>2)

  %
  % things to do always when SFA_STRUCTS{hdl}.step is either 'preprocessing' or 'expansion' 
  % (no matter whether it is invoked for the first time or once again)
  %
  switch SFA_STRUCTS{hdl}.step,
     case 'preprocessing',
        if strncmpi(SFA_STRUCTS{hdl}.pp_type, 'PCA',3),     % either 'PCA','PCA2' or 'PCAVAR'
          lcov_update(SFA_STRUCTS{hdl}.lcov_hdl, arg);
        else %SFA1
          sfa_step(SFA_STRUCTS{hdl}.sfa1_hdl, arg, 'preprocessing');
        end

     case 'expansion',
        if isempty(arg), return, end
        % project data and reduce dimensionality
        arg=arg-repmat(SFA_STRUCTS{hdl}.avg0, size(arg,1), 1);
        arg=expansion(hdl, arg*SFA_STRUCTS{hdl}.W0');
        lcov_update(SFA_STRUCTS{hdl}.xp_hdl, arg);

        if strcmp(method,'TIMESERIES')
            lcov_update(SFA_STRUCTS{hdl}.diff_hdl, timediff(arg, SFA_STRUCTS{hdl}.ax_type));
        elseif strcmp(method,'CLASSIF')
            % extension /WK/08/2009: generate the difference of all pattern
            % pairs in 'pdiff'
            K = size(arg,1);
            pdiff = [];
            for k=1:K-1
                pdiff = [pdiff; repmat(arg(k,:),K-k,1) - arg(k+1:K,:)];
            end
            lcov_update(SFA_STRUCTS{hdl}.diff_hdl, pdiff);
        else
            warning(['method=' method ' not allowed in expansion step']);
        end

  end
  